@@ -53,20 +53,8 @@
 
 #include "xkbcomp-priv.h"
 #include "ast-build.h"
-#include "parser-priv.h"
 #include "include.h"
 
-ATTR_MALLOC static void *
-malloc_or_die(size_t size)
-{
-    void *p = malloc(size);
-    if (!p) {
-        fprintf(stderr, "Out of memory\n");
-        exit(1);
-    }
-    return p;
-}
-
 ParseCommon *
 AppendStmt(ParseCommon *to, ParseCommon *append)
 {
@@ -81,17 +69,63 @@ AppendStmt(ParseCommon *to, ParseCommon *append)
     return to;
 }
 
-ExprDef *
-ExprCreate(enum expr_op_type op, enum expr_value_type type)
+static ExprDef *
+ExprCreate(enum expr_op_type op, enum expr_value_type type, size_t size)
 {
-    ExprDef *expr;
-
-    expr = malloc_or_die(sizeof(*expr));
+    ExprDef *expr = malloc(size);
+    if (!expr)
+        return NULL;
 
     expr->common.type = STMT_EXPR;
     expr->common.next = NULL;
-    expr->op = op;
-    expr->value_type = type;
+    expr->expr.op = op;
+    expr->expr.value_type = type;
+
+    return expr;
+}
+
+#define EXPR_CREATE(type_, name_, op_, value_type_) \
+    ExprDef *name_ = ExprCreate(op_, value_type_, sizeof(type_)); \
+    if (!name_) \
+        return NULL;
+
+ExprDef *
+ExprCreateString(xkb_atom_t str)
+{
+    EXPR_CREATE(ExprString, expr, EXPR_VALUE, EXPR_TYPE_STRING);
+    expr->string.str = str;
+    return expr;
+}
+
+ExprDef *
+ExprCreateInteger(int ival)
+{
+    EXPR_CREATE(ExprInteger, expr, EXPR_VALUE, EXPR_TYPE_INT);
+    expr->integer.ival = ival;
+    return expr;
+}
+
+ExprDef *
+ExprCreateBoolean(bool set)
+{
+    EXPR_CREATE(ExprBoolean, expr, EXPR_VALUE, EXPR_TYPE_BOOLEAN);
+    expr->boolean.set = set;
+    return expr;
+}
+
+ExprDef *
+ExprCreateKeyName(xkb_atom_t key_name)
+{
+    EXPR_CREATE(ExprKeyName, expr, EXPR_VALUE, EXPR_TYPE_KEYNAME);
+    expr->key_name.key_name = key_name;
+    return expr;
+}
+
+ExprDef *
+ExprCreateIdent(xkb_atom_t ident)
+{
+    EXPR_CREATE(ExprIdent, expr, EXPR_IDENT, EXPR_TYPE_UNKNOWN);
+    expr->ident.ident = ident;
     return expr;
 }
 
@@ -99,287 +133,306 @@ ExprDef *
 ExprCreateUnary(enum expr_op_type op, enum expr_value_type type,
                 ExprDef *child)
 {
-    ExprDef *expr;
-    expr = malloc_or_die(sizeof(*expr));
-
-    expr->common.type = STMT_EXPR;
-    expr->common.next = NULL;
-    expr->op = op;
-    expr->value_type = type;
-    expr->value.child = child;
+    EXPR_CREATE(ExprUnary, expr, op, type);
+    expr->unary.child = child;
     return expr;
 }
 
 ExprDef *
 ExprCreateBinary(enum expr_op_type op, ExprDef *left, ExprDef *right)
 {
-    ExprDef *expr;
+    EXPR_CREATE(ExprBinary, expr, op, EXPR_TYPE_UNKNOWN);
 
-    expr = malloc_or_die(sizeof(*expr));
+    if (op == EXPR_ASSIGN || left->expr.value_type == EXPR_TYPE_UNKNOWN)
+        expr->expr.value_type = right->expr.value_type;
+    else if (left->expr.value_type == right->expr.value_type ||
+             right->expr.value_type == EXPR_TYPE_UNKNOWN)
+        expr->expr.value_type = left->expr.value_type;
+    expr->binary.left = left;
+    expr->binary.right = right;
+
+    return expr;
+}
+
+ExprDef *
+ExprCreateFieldRef(xkb_atom_t element, xkb_atom_t field)
+{
+    EXPR_CREATE(ExprFieldRef, expr, EXPR_FIELD_REF, EXPR_TYPE_UNKNOWN);
+    expr->field_ref.element = element;
+    expr->field_ref.field = field;
+    return expr;
+}
+
+ExprDef *
+ExprCreateArrayRef(xkb_atom_t element, xkb_atom_t field, ExprDef *entry)
+{
+    EXPR_CREATE(ExprArrayRef, expr, EXPR_ARRAY_REF, EXPR_TYPE_UNKNOWN);
+    expr->array_ref.element = element;
+    expr->array_ref.field = field;
+    expr->array_ref.entry = entry;
+    return expr;
+}
+
+ExprDef *
+ExprCreateAction(xkb_atom_t name, ExprDef *args)
+{
+    EXPR_CREATE(ExprAction, expr, EXPR_ACTION_DECL, EXPR_TYPE_UNKNOWN);
+    expr->action.name = name;
+    expr->action.args = args;
+    return expr;
+}
+
+ExprDef *
+ExprCreateKeysymList(xkb_keysym_t sym)
+{
+    EXPR_CREATE(ExprKeysymList, expr, EXPR_KEYSYM_LIST, EXPR_TYPE_SYMBOLS);
+
+    darray_init(expr->keysym_list.syms);
+    darray_init(expr->keysym_list.symsMapIndex);
+    darray_init(expr->keysym_list.symsNumEntries);
+
+    darray_append(expr->keysym_list.syms, sym);
+    darray_append(expr->keysym_list.symsMapIndex, 0);
+    darray_append(expr->keysym_list.symsNumEntries, 1);
+
+    return expr;
+}
+
+ExprDef *
+ExprCreateMultiKeysymList(ExprDef *expr)
+{
+    unsigned nLevels = darray_size(expr->keysym_list.symsMapIndex);
+
+    darray_resize(expr->keysym_list.symsMapIndex, 1);
+    darray_resize(expr->keysym_list.symsNumEntries, 1);
+    darray_item(expr->keysym_list.symsMapIndex, 0) = 0;
+    darray_item(expr->keysym_list.symsNumEntries, 0) = nLevels;
+
+    return expr;
+}
+
+ExprDef *
+ExprAppendKeysymList(ExprDef *expr, xkb_keysym_t sym)
+{
+    unsigned nSyms = darray_size(expr->keysym_list.syms);
+
+    darray_append(expr->keysym_list.symsMapIndex, nSyms);
+    darray_append(expr->keysym_list.symsNumEntries, 1);
+    darray_append(expr->keysym_list.syms, sym);
+
+    return expr;
+}
+
+ExprDef *
+ExprAppendMultiKeysymList(ExprDef *expr, ExprDef *append)
+{
+    xkb_keysym_t *syms;
+    unsigned nSyms = darray_size(expr->keysym_list.syms);
+    unsigned numEntries = darray_size(append->keysym_list.syms);
+
+    darray_append(expr->keysym_list.symsMapIndex, nSyms);
+    darray_append(expr->keysym_list.symsNumEntries, numEntries);
+    darray_steal(append->keysym_list.syms, &syms, NULL);
+    darray_append_items(expr->keysym_list.syms, syms, numEntries);
+
+    FreeStmt((ParseCommon *) &append);
 
-    expr->common.type = STMT_EXPR;
-    expr->common.next = NULL;
-    expr->op = op;
-    if (op == EXPR_ASSIGN || left->value_type == EXPR_TYPE_UNKNOWN)
-        expr->value_type = right->value_type;
-    else if (left->value_type == right->value_type ||
-             right->value_type == EXPR_TYPE_UNKNOWN)
-        expr->value_type = left->value_type;
-    else
-        expr->value_type = EXPR_TYPE_UNKNOWN;
-    expr->value.binary.left = left;
-    expr->value.binary.right = right;
     return expr;
 }
 
 KeycodeDef *
 KeycodeCreate(xkb_atom_t name, int64_t value)
 {
-    KeycodeDef *def;
-
-    def = malloc_or_die(sizeof(*def));
+    KeycodeDef *def = malloc(sizeof(*def));
+    if (!def)
+        return NULL;
 
     def->common.type = STMT_KEYCODE;
     def->common.next = NULL;
     def->name = name;
     def->value = value;
+
     return def;
 }
 
 KeyAliasDef *
 KeyAliasCreate(xkb_atom_t alias, xkb_atom_t real)
 {
-    KeyAliasDef *def;
-
-    def = malloc_or_die(sizeof(*def));
+    KeyAliasDef *def = malloc(sizeof(*def));
+    if (!def)
+        return NULL;
 
     def->common.type = STMT_ALIAS;
     def->common.next = NULL;
     def->alias = alias;
     def->real = real;
+
     return def;
 }
 
 VModDef *
-VModCreate(xkb_atom_t name, ExprDef * value)
+VModCreate(xkb_atom_t name, ExprDef *value)
 {
-    VModDef *def;
-
-    def = malloc_or_die(sizeof(*def));
+    VModDef *def = malloc(sizeof(*def));
+    if (!def)
+        return NULL;
 
     def->common.type = STMT_VMOD;
     def->common.next = NULL;
     def->name = name;
     def->value = value;
+
     return def;
 }
 
 VarDef *
-VarCreate(ExprDef * name, ExprDef * value)
+VarCreate(ExprDef *name, ExprDef *value)
 {
-    VarDef *def;
-    def = malloc_or_die(sizeof(*def));
+    VarDef *def = malloc(sizeof(*def));
+    if (!def)
+        return NULL;
 
     def->common.type = STMT_VAR;
     def->common.next = NULL;
     def->name = name;
     def->value = value;
+
     return def;
 }
 
 VarDef *
-BoolVarCreate(xkb_atom_t nameToken, unsigned set)
+BoolVarCreate(xkb_atom_t ident, bool set)
 {
     ExprDef *name, *value;
-
-    name = ExprCreate(EXPR_IDENT, EXPR_TYPE_UNKNOWN);
-    name->value.str = nameToken;
-    value = ExprCreate(EXPR_VALUE, EXPR_TYPE_BOOLEAN);
-    value->value.uval = set;
-    return VarCreate(name, value);
+    VarDef *def;
+    if (!(name = ExprCreateIdent(ident))) {
+        return NULL;
+    }
+    if (!(value = ExprCreateBoolean(set))) {
+        FreeStmt((ParseCommon *) name);
+        return NULL;
+    }
+    if (!(def = VarCreate(name, value))) {
+        FreeStmt((ParseCommon *) name);
+        FreeStmt((ParseCommon *) value);
+        return NULL;
+    }
+    return def;
 }
 
 InterpDef *
-InterpCreate(char *sym, ExprDef * match)
+InterpCreate(xkb_keysym_t sym, ExprDef *match)
 {
-    InterpDef *def;
-
-    def = malloc_or_die(sizeof(*def));
+    InterpDef *def = malloc(sizeof(*def));
+    if (!def)
+        return NULL;
 
     def->common.type = STMT_INTERP;
     def->common.next = NULL;
     def->sym = sym;
     def->match = match;
+    def->def = NULL;
+
     return def;
 }
 
 KeyTypeDef *
-KeyTypeCreate(xkb_atom_t name, VarDef * body)
+KeyTypeCreate(xkb_atom_t name, VarDef *body)
 {
-    KeyTypeDef *def;
-
-    def = malloc_or_die(sizeof(*def));
+    KeyTypeDef *def = malloc(sizeof(*def));
+    if (!def)
+        return NULL;
 
     def->common.type = STMT_TYPE;
     def->common.next = NULL;
     def->merge = MERGE_DEFAULT;
     def->name = name;
     def->body = body;
+
     return def;
 }
 
 SymbolsDef *
-SymbolsCreate(xkb_atom_t keyName, ExprDef *symbols)
+SymbolsCreate(xkb_atom_t keyName, VarDef *symbols)
 {
-    SymbolsDef *def;
-
-    def = malloc_or_die(sizeof(*def));
+    SymbolsDef *def = malloc(sizeof(*def));
+    if (!def)
+        return NULL;
 
     def->common.type = STMT_SYMBOLS;
     def->common.next = NULL;
     def->merge = MERGE_DEFAULT;
     def->keyName = keyName;
     def->symbols = symbols;
+
     return def;
 }
 
 GroupCompatDef *
-GroupCompatCreate(int group, ExprDef * val)
+GroupCompatCreate(unsigned group, ExprDef *val)
 {
-    GroupCompatDef *def;
-
-    def = malloc_or_die(sizeof(*def));
+    GroupCompatDef *def = malloc(sizeof(*def));
+    if (!def)
+        return NULL;
 
     def->common.type = STMT_GROUP_COMPAT;
     def->common.next = NULL;
     def->merge = MERGE_DEFAULT;
     def->group = group;
     def->def = val;
+
     return def;
 }
 
 ModMapDef *
-ModMapCreate(uint32_t modifier, ExprDef * keys)
+ModMapCreate(xkb_atom_t modifier, ExprDef *keys)
 {
-    ModMapDef *def;
-
-    def = malloc_or_die(sizeof(*def));
+    ModMapDef *def = malloc(sizeof(*def));
+    if (!def)
+        return NULL;
 
     def->common.type = STMT_MODMAP;
     def->common.next = NULL;
     def->merge = MERGE_DEFAULT;
     def->modifier = modifier;
     def->keys = keys;
+
     return def;
 }
 
-IndicatorMapDef *
-IndicatorMapCreate(xkb_atom_t name, VarDef * body)
+LedMapDef *
+LedMapCreate(xkb_atom_t name, VarDef *body)
 {
-    IndicatorMapDef *def;
+    LedMapDef *def = malloc(sizeof(*def));
+    if (!def)
+        return NULL;
 
-    def = malloc_or_die(sizeof(*def));
-
-    def->common.type = STMT_INDICATOR_MAP;
+    def->common.type = STMT_LED_MAP;
     def->common.next = NULL;
     def->merge = MERGE_DEFAULT;
     def->name = name;
     def->body = body;
+
     return def;
 }
 
-IndicatorNameDef *
-IndicatorNameCreate(int ndx, ExprDef * name, bool virtual)
+LedNameDef *
+LedNameCreate(unsigned ndx, ExprDef *name, bool virtual)
 {
-    IndicatorNameDef *def;
+    LedNameDef *def = malloc(sizeof(*def));
+    if (!def)
+        return NULL;
 
-    def = malloc_or_die(sizeof(*def));
-
-    def->common.type = STMT_INDICATOR_NAME;
+    def->common.type = STMT_LED_NAME;
     def->common.next = NULL;
     def->merge = MERGE_DEFAULT;
     def->ndx = ndx;
     def->name = name;
     def->virtual = virtual;
-    return def;
-}
-
-ExprDef *
-ActionCreate(xkb_atom_t name, ExprDef * args)
-{
-    ExprDef *act;
-
-    act = malloc_or_die(sizeof(*act));
-
-    act->common.type = STMT_EXPR;
-    act->common.next = NULL;
-    act->op = EXPR_ACTION_DECL;
-    act->value.action.name = name;
-    act->value.action.args = args;
-    return act;
-}
-
-ExprDef *
-CreateKeysymList(char *sym)
-{
-    ExprDef *def;
-
-    def = ExprCreate(EXPR_KEYSYM_LIST, EXPR_TYPE_SYMBOLS);
-
-    darray_init(def->value.list.syms);
-    darray_init(def->value.list.symsMapIndex);
-    darray_init(def->value.list.symsNumEntries);
-
-    darray_append(def->value.list.syms, sym);
-    darray_append(def->value.list.symsMapIndex, 0);
-    darray_append(def->value.list.symsNumEntries, 1);
 
     return def;
 }
 
-ExprDef *
-CreateMultiKeysymList(ExprDef *list)
-{
-    size_t nLevels = darray_size(list->value.list.symsMapIndex);
-
-    darray_resize(list->value.list.symsMapIndex, 1);
-    darray_resize(list->value.list.symsNumEntries, 1);
-    darray_item(list->value.list.symsMapIndex, 0) = 0;
-    darray_item(list->value.list.symsNumEntries, 0) = nLevels;
-
-    return list;
-}
-
-ExprDef *
-AppendKeysymList(ExprDef * list, char *sym)
-{
-    size_t nSyms = darray_size(list->value.list.syms);
-
-    darray_append(list->value.list.symsMapIndex, nSyms);
-    darray_append(list->value.list.symsNumEntries, 1);
-    darray_append(list->value.list.syms, sym);
-
-    return list;
-}
-
-ExprDef *
-AppendMultiKeysymList(ExprDef * list, ExprDef * append)
-{
-    size_t nSyms = darray_size(list->value.list.syms);
-    size_t numEntries = darray_size(append->value.list.syms);
-
-    darray_append(list->value.list.symsMapIndex, nSyms);
-    darray_append(list->value.list.symsNumEntries, numEntries);
-    darray_append_items(list->value.list.syms,
-                        darray_mem(append->value.list.syms, 0),
-                        numEntries);
-
-    darray_resize(append->value.list.syms, 0);
-    FreeStmt(&append->common);
-
-    return list;
-}
-
 static void
 FreeInclude(IncludeStmt *incl);
 
@@ -455,33 +508,9 @@ err:
     return NULL;
 }
 
-/*
- * All latin-1 alphanumerics, plus parens, slash, minus, underscore and
- * wildcards.
- */
-static const unsigned char componentSpecLegal[] = {
-    0x00, 0x00, 0x00, 0x00, 0x00, 0xa7, 0xff, 0x83,
-    0xfe, 0xff, 0xff, 0x87, 0xfe, 0xff, 0xff, 0x07,
-    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
-    0xff, 0xff, 0x7f, 0xff, 0xff, 0xff, 0x7f, 0xff
-};
-
-static void
-EnsureSafeMapName(char *name)
-{
-    if (!name)
-        return;
-
-    while (*name != '\0') {
-        if ((componentSpecLegal[(*name) / 8] & (1 << ((*name) % 8))) == 0)
-            *name = '_';
-        name++;
-    }
-}
-
 XkbFile *
-XkbFileCreate(struct xkb_context *ctx, enum xkb_file_type type, char *name,
-              ParseCommon *defs, enum xkb_map_flags flags)
+XkbFileCreate(enum xkb_file_type type, char *name, ParseCommon *defs,
+              enum xkb_map_flags flags)
 {
     XkbFile *file;
 
@@ -489,61 +518,67 @@ XkbFileCreate(struct xkb_context *ctx, enum xkb_file_type type, char *name,
     if (!file)
         return NULL;
 
-    EnsureSafeMapName(name);
+    XkbEscapeMapName(name);
     file->file_type = type;
     file->topName = strdup_safe(name);
     file->name = name;
     file->defs = defs;
-    file->id = xkb_context_take_file_id(ctx);
     file->flags = flags;
+
     return file;
 }
 
 XkbFile *
 XkbFileFromComponents(struct xkb_context *ctx,
-                      struct xkb_component_names *kkctgs)
+                      const struct xkb_component_names *kkctgs)
 {
-    IncludeStmt *inc;
-    XkbFile *keycodes, *types, *compat, *symbols;
+    char *const components[] = {
+        kkctgs->keycodes, kkctgs->types,
+        kkctgs->compat, kkctgs->symbols,
+    };
+    enum xkb_file_type type;
+    IncludeStmt *include = NULL;
+    XkbFile *file = NULL;
+    ParseCommon *defs = NULL;
 
-    inc = IncludeCreate(ctx, kkctgs->keycodes, MERGE_DEFAULT);
-    keycodes = XkbFileCreate(ctx, FILE_TYPE_KEYCODES, NULL,
-                             (ParseCommon *) inc, 0);
+    for (type = FIRST_KEYMAP_FILE_TYPE; type <= LAST_KEYMAP_FILE_TYPE; type++) {
+        include = IncludeCreate(ctx, components[type], MERGE_DEFAULT);
+        if (!include)
+            goto err;
 
-    inc = IncludeCreate(ctx, kkctgs->types, MERGE_DEFAULT);
-    types = XkbFileCreate(ctx, FILE_TYPE_TYPES, NULL,
-                          (ParseCommon *) inc, 0);
-    AppendStmt(&keycodes->common, &types->common);
+        file = XkbFileCreate(type, NULL, (ParseCommon *) include, 0);
+        if (!file) {
+            FreeInclude(include);
+            goto err;
+        }
 
-    inc = IncludeCreate(ctx, kkctgs->compat, MERGE_DEFAULT);
-    compat = XkbFileCreate(ctx, FILE_TYPE_COMPAT, NULL,
-                           (ParseCommon *) inc, 0);
-    AppendStmt(&keycodes->common, &compat->common);
+        defs = AppendStmt(defs, &file->common);
+    }
 
-    inc = IncludeCreate(ctx, kkctgs->symbols, MERGE_DEFAULT);
-    symbols = XkbFileCreate(ctx, FILE_TYPE_SYMBOLS, NULL,
-                            (ParseCommon *) inc, 0);
-    AppendStmt(&keycodes->common, &symbols->common);
+    file = XkbFileCreate(FILE_TYPE_KEYMAP, NULL, defs, 0);
+    if (!file)
+        goto err;
 
-    return XkbFileCreate(ctx, FILE_TYPE_KEYMAP, NULL,
-                         &keycodes->common, 0);
+    return file;
+
+err:
+    FreeXkbFile((XkbFile *) defs);
+    return NULL;
 }
 
 static void
 FreeExpr(ExprDef *expr)
 {
-    char **sym;
-
     if (!expr)
         return;
 
-    switch (expr->op) {
+    switch (expr->expr.op) {
     case EXPR_ACTION_LIST:
     case EXPR_NEGATE:
     case EXPR_UNARY_PLUS:
     case EXPR_NOT:
     case EXPR_INVERT:
-        FreeStmt(&expr->value.child->common);
+        FreeStmt((ParseCommon *) expr->unary.child);
         break;
 
     case EXPR_DIVIDE:
@@ -551,24 +586,22 @@ FreeExpr(ExprDef *expr)
     case EXPR_SUBTRACT:
     case EXPR_MULTIPLY:
     case EXPR_ASSIGN:
-        FreeStmt(&expr->value.binary.left->common);
-        FreeStmt(&expr->value.binary.right->common);
+        FreeStmt((ParseCommon *) expr->binary.left);
+        FreeStmt((ParseCommon *) expr->binary.right);
         break;
 
     case EXPR_ACTION_DECL:
-        FreeStmt(&expr->value.action.args->common);
+        FreeStmt((ParseCommon *) expr->action.args);
         break;
 
     case EXPR_ARRAY_REF:
-        FreeStmt(&expr->value.array.entry->common);
+        FreeStmt((ParseCommon *) expr->array_ref.entry);
         break;
 
     case EXPR_KEYSYM_LIST:
-        darray_foreach(sym, expr->value.list.syms)
-            free(*sym);
-        darray_free(expr->value.list.syms);
-        darray_free(expr->value.list.symsMapIndex);
-        darray_free(expr->value.list.symsNumEntries);
+        darray_free(expr->keysym_list.syms);
+        darray_free(expr->keysym_list.symsMapIndex);
+        darray_free(expr->keysym_list.symsNumEntries);
         break;
 
     default:
@@ -599,12 +632,10 @@ void
 FreeStmt(ParseCommon *stmt)
 {
     ParseCommon *next;
-    YYSTYPE u;
 
     while (stmt)
     {
         next = stmt->next;
-        u.any = stmt;
 
         switch (stmt->type) {
         case STMT_INCLUDE:
@@ -613,37 +644,36 @@ FreeStmt(ParseCommon *stmt)
             stmt = NULL;
             break;
         case STMT_EXPR:
-            FreeExpr(u.expr);
+            FreeExpr((ExprDef *) stmt);
             break;
         case STMT_VAR:
-            FreeStmt(&u.var->name->common);
-            FreeStmt(&u.var->value->common);
+            FreeStmt((ParseCommon *) ((VarDef *) stmt)->name);
+            FreeStmt((ParseCommon *) ((VarDef *) stmt)->value);
             break;
         case STMT_TYPE:
-            FreeStmt(&u.keyType->body->common);
+            FreeStmt((ParseCommon *) ((KeyTypeDef *) stmt)->body);
             break;
         case STMT_INTERP:
-            free(u.interp->sym);
-            FreeStmt(&u.interp->match->common);
-            FreeStmt(&u.interp->def->common);
+            FreeStmt((ParseCommon *) ((InterpDef *) stmt)->match);
+            FreeStmt((ParseCommon *) ((InterpDef *) stmt)->def);
             break;
         case STMT_VMOD:
-            FreeStmt(&u.vmod->value->common);
+            FreeStmt((ParseCommon *) ((VModDef *) stmt)->value);
             break;
         case STMT_SYMBOLS:
-            FreeStmt(&u.syms->symbols->common);
+            FreeStmt((ParseCommon *) ((SymbolsDef *) stmt)->symbols);
             break;
         case STMT_MODMAP:
-            FreeStmt(&u.modMask->keys->common);
+            FreeStmt((ParseCommon *) ((ModMapDef *) stmt)->keys);
             break;
         case STMT_GROUP_COMPAT:
-            FreeStmt(&u.groupCompat->def->common);
+            FreeStmt((ParseCommon *) ((GroupCompatDef *) stmt)->def);
             break;
-        case STMT_INDICATOR_MAP:
-            FreeStmt(&u.ledMap->body->common);
+        case STMT_LED_MAP:
+            FreeStmt((ParseCommon *) ((LedMapDef *) stmt)->body);
             break;
-        case STMT_INDICATOR_NAME:
-            FreeStmt(&u.ledName->name->common);
+        case STMT_LED_NAME:
+            FreeStmt((ParseCommon *) ((LedNameDef *) stmt)->name);
             break;
         default:
             break;
@@ -692,6 +722,7 @@ static const char *xkb_file_type_strings[_FILE_TYPE_NUM_ENTRIES] = {
     [FILE_TYPE_TYPES] = "xkb_types",
     [FILE_TYPE_COMPAT] = "xkb_compatibility",
     [FILE_TYPE_SYMBOLS] = "xkb_symbols",
+    [FILE_TYPE_GEOMETRY] = "xkb_geometry",
     [FILE_TYPE_KEYMAP] = "xkb_keymap",
     [FILE_TYPE_RULES] = "rules",
 };
@@ -717,8 +748,8 @@ static const char *stmt_type_strings[_STMT_NUM_VALUES] = {
     [STMT_SYMBOLS] = "key symbols definition",
     [STMT_MODMAP] = "modifier map declaration",
     [STMT_GROUP_COMPAT] = "group declaration",
-    [STMT_INDICATOR_MAP] = "indicator map declaration",
-    [STMT_INDICATOR_NAME] = "indicator name declaration",
+    [STMT_LED_MAP] = "indicator map declaration",
+    [STMT_LED_NAME] = "indicator name declaration",
 };
 
 const char *
