@@ -37,26 +37,26 @@ ExprResolveLhs(struct xkb_context *ctx, const ExprDef *expr,
                const char **elem_rtrn, const char **field_rtrn,
                ExprDef **index_rtrn)
 {
-    switch (expr->op) {
+    switch (expr->expr.op) {
     case EXPR_IDENT:
         *elem_rtrn = NULL;
-        *field_rtrn = xkb_atom_text(ctx, expr->value.str);
+        *field_rtrn = xkb_atom_text(ctx, expr->ident.ident);
         *index_rtrn = NULL;
         return true;
     case EXPR_FIELD_REF:
-        *elem_rtrn = xkb_atom_text(ctx, expr->value.field.element);
-        *field_rtrn = xkb_atom_text(ctx, expr->value.field.field);
+        *elem_rtrn = xkb_atom_text(ctx, expr->field_ref.element);
+        *field_rtrn = xkb_atom_text(ctx, expr->field_ref.field);
         *index_rtrn = NULL;
         return true;
     case EXPR_ARRAY_REF:
-        *elem_rtrn = xkb_atom_text(ctx, expr->value.array.element);
-        *field_rtrn = xkb_atom_text(ctx, expr->value.array.field);
-        *index_rtrn = expr->value.array.entry;
+        *elem_rtrn = xkb_atom_text(ctx, expr->array_ref.element);
+        *field_rtrn = xkb_atom_text(ctx, expr->array_ref.field);
+        *index_rtrn = expr->array_ref.entry;
         return true;
     default:
         break;
     }
-    log_wsgo(ctx, "Unexpected operator %d in ResolveLhs\n", expr->op);
+    log_wsgo(ctx, "Unexpected operator %d in ResolveLhs\n", expr->expr.op);
     return false;
 }
 
@@ -83,7 +83,7 @@ SimpleLookup(struct xkb_context *ctx, const void *priv, xkb_atom_t field,
 
 /* Data passed in the *priv argument for LookupModMask. */
 typedef struct {
-    const struct xkb_keymap *keymap;
+    const struct xkb_mod_set *mods;
     enum mod_type mod_type;
 } LookupModMaskPriv;
 
@@ -94,7 +94,7 @@ LookupModMask(struct xkb_context *ctx, const void *priv, xkb_atom_t field,
     const char *str;
     xkb_mod_index_t ndx;
     const LookupModMaskPriv *arg = priv;
-    const struct xkb_keymap *keymap = arg->keymap;
+    const struct xkb_mod_set *mods = arg->mods;
     enum mod_type mod_type = arg->mod_type;
 
     if (type != EXPR_TYPE_INT)
@@ -112,11 +112,11 @@ LookupModMask(struct xkb_context *ctx, const void *priv, xkb_atom_t field,
         return true;
     }
 
-    ndx = ModNameToIndex(keymap, field, mod_type);
+    ndx = XkbModNameToIndex(mods, field, mod_type);
     if (ndx == XKB_MOD_INVALID)
         return false;
 
-    *val_rtrn = (1 << ndx);
+    *val_rtrn = (1u << ndx);
     return true;
 }
 
@@ -127,19 +127,19 @@ ExprResolveBoolean(struct xkb_context *ctx, const ExprDef *expr,
     bool ok = false;
     const char *ident;
 
-    switch (expr->op) {
+    switch (expr->expr.op) {
     case EXPR_VALUE:
-        if (expr->value_type != EXPR_TYPE_BOOLEAN) {
+        if (expr->expr.value_type != EXPR_TYPE_BOOLEAN) {
             log_err(ctx,
                     "Found constant of type %s where boolean was expected\n",
-                    expr_value_type_to_string(expr->value_type));
+                    expr_value_type_to_string(expr->expr.value_type));
             return false;
         }
-        *set_rtrn = !!expr->value.ival;
+        *set_rtrn = expr->boolean.set;
         return true;
 
     case EXPR_IDENT:
-        ident = xkb_atom_text(ctx, expr->value.str);
+        ident = xkb_atom_text(ctx, expr->ident.ident);
         if (ident) {
             if (istreq(ident, "true") ||
                 istreq(ident, "yes") ||
@@ -154,14 +154,13 @@ ExprResolveBoolean(struct xkb_context *ctx, const ExprDef *expr,
                 return true;
             }
         }
-        log_err(ctx, "Identifier \"%s\" of type boolean is unknown\n",
-                xkb_atom_text(ctx, expr->value.str));
+        log_err(ctx, "Identifier \"%s\" of type boolean is unknown\n", ident);
         return false;
 
     case EXPR_FIELD_REF:
         log_err(ctx, "Default \"%s.%s\" of type boolean is unknown\n",
-                xkb_atom_text(ctx, expr->value.field.element),
-                xkb_atom_text(ctx, expr->value.field.field));
+                xkb_atom_text(ctx, expr->field_ref.element),
+                xkb_atom_text(ctx, expr->field_ref.field));
         return false;
 
     case EXPR_INVERT:
@@ -178,11 +177,12 @@ ExprResolveBoolean(struct xkb_context *ctx, const ExprDef *expr,
     case EXPR_NEGATE:
     case EXPR_UNARY_PLUS:
         log_err(ctx, "%s of boolean values not permitted\n",
-                expr_op_type_to_string(expr->op));
+                expr_op_type_to_string(expr->expr.op));
         break;
 
     default:
-        log_wsgo(ctx, "Unknown operator %d in ResolveBoolean\n", expr->op);
+        log_wsgo(ctx, "Unknown operator %d in ResolveBoolean\n",
+                 expr->expr.op);
         break;
     }
 
@@ -194,32 +194,28 @@ ExprResolveKeyCode(struct xkb_context *ctx, const ExprDef *expr,
                    xkb_keycode_t *kc)
 {
     xkb_keycode_t leftRtrn, rightRtrn;
-    ExprDef *left, *right;
 
-    switch (expr->op) {
+    switch (expr->expr.op) {
     case EXPR_VALUE:
-        if (expr->value_type != EXPR_TYPE_INT) {
+        if (expr->expr.value_type != EXPR_TYPE_INT) {
             log_err(ctx,
                     "Found constant of type %s where an int was expected\n",
-                    expr_value_type_to_string(expr->value_type));
+                    expr_value_type_to_string(expr->expr.value_type));
             return false;
         }
 
-        *kc = expr->value.uval;
+        *kc = (xkb_keycode_t) expr->integer.ival;
         return true;
 
     case EXPR_ADD:
     case EXPR_SUBTRACT:
     case EXPR_MULTIPLY:
     case EXPR_DIVIDE:
-        left = expr->value.binary.left;
-        right = expr->value.binary.right;
-
-        if (!ExprResolveKeyCode(ctx, left, &leftRtrn) ||
-            !ExprResolveKeyCode(ctx, right, &rightRtrn))
+        if (!ExprResolveKeyCode(ctx, expr->binary.left, &leftRtrn) ||
+            !ExprResolveKeyCode(ctx, expr->binary.right, &rightRtrn))
             return false;
 
-        switch (expr->op) {
+        switch (expr->expr.op) {
         case EXPR_ADD:
             *kc = leftRtrn + rightRtrn;
             break;
@@ -245,19 +241,18 @@ ExprResolveKeyCode(struct xkb_context *ctx, const ExprDef *expr,
         return true;
 
     case EXPR_NEGATE:
-        left = expr->value.child;
-        if (!ExprResolveKeyCode(ctx, left, &leftRtrn))
+        if (!ExprResolveKeyCode(ctx, expr->unary.child, &leftRtrn))
             return false;
 
         *kc = ~leftRtrn;
         return true;
 
     case EXPR_UNARY_PLUS:
-        left = expr->value.child;
-        return ExprResolveKeyCode(ctx, left, kc);
+        return ExprResolveKeyCode(ctx, expr->unary.child, kc);
 
     default:
-        log_wsgo(ctx, "Unknown operator %d in ResolveKeyCode\n", expr->op);
+        log_wsgo(ctx, "Unknown operator %d in ResolveKeyCode\n",
+                 expr->expr.op);
         break;
     }
 
@@ -284,25 +279,25 @@ ExprResolveIntegerLookup(struct xkb_context *ctx, const ExprDef *expr,
     unsigned u;
     ExprDef *left, *right;
 
-    switch (expr->op) {
+    switch (expr->expr.op) {
     case EXPR_VALUE:
-        if (expr->value_type != EXPR_TYPE_INT) {
+        if (expr->expr.value_type != EXPR_TYPE_INT) {
             log_err(ctx,
                     "Found constant of type %s where an int was expected\n",
-                    expr_value_type_to_string(expr->value_type));
+                    expr_value_type_to_string(expr->expr.value_type));
             return false;
         }
 
-        *val_rtrn = expr->value.ival;
+        *val_rtrn = expr->integer.ival;
         return true;
 
     case EXPR_IDENT:
         if (lookup)
-            ok = lookup(ctx, lookupPriv, expr->value.str, EXPR_TYPE_INT, &u);
+            ok = lookup(ctx, lookupPriv, expr->ident.ident, EXPR_TYPE_INT, &u);
 
         if (!ok)
             log_err(ctx, "Identifier \"%s\" of type int is unknown\n",
-                    xkb_atom_text(ctx, expr->value.str));
+                    xkb_atom_text(ctx, expr->ident.ident));
         else
             *val_rtrn = (int) u;
 
@@ -310,21 +305,21 @@ ExprResolveIntegerLookup(struct xkb_context *ctx, const ExprDef *expr,
 
     case EXPR_FIELD_REF:
         log_err(ctx, "Default \"%s.%s\" of type int is unknown\n",
-                xkb_atom_text(ctx, expr->value.field.element),
-                xkb_atom_text(ctx, expr->value.field.field));
+                xkb_atom_text(ctx, expr->field_ref.element),
+                xkb_atom_text(ctx, expr->field_ref.field));
         return false;
 
     case EXPR_ADD:
     case EXPR_SUBTRACT:
     case EXPR_MULTIPLY:
     case EXPR_DIVIDE:
-        left = expr->value.binary.left;
-        right = expr->value.binary.right;
+        left = expr->binary.left;
+        right = expr->binary.right;
         if (!ExprResolveIntegerLookup(ctx, left, &l, lookup, lookupPriv) ||
             !ExprResolveIntegerLookup(ctx, right, &r, lookup, lookupPriv))
             return false;
 
-        switch (expr->op) {
+        switch (expr->expr.op) {
         case EXPR_ADD:
             *val_rtrn = l + r;
             break;
@@ -342,7 +337,9 @@ ExprResolveIntegerLookup(struct xkb_context *ctx, const ExprDef *expr,
             *val_rtrn = l / r;
             break;
         default:
-            break;
+            log_err(ctx, "%s of integers not permitted\n",
+                    expr_op_type_to_string(expr->expr.op));
+            return false;
         }
 
         return true;
@@ -357,20 +354,21 @@ ExprResolveIntegerLookup(struct xkb_context *ctx, const ExprDef *expr,
 
     case EXPR_INVERT:
     case EXPR_NEGATE:
-        left = expr->value.child;
+        left = expr->unary.child;
         if (!ExprResolveIntegerLookup(ctx, left, &l, lookup, lookupPriv))
             return false;
 
-        *val_rtrn = (expr->op == EXPR_NEGATE ? -l : ~l);
+        *val_rtrn = (expr->expr.op == EXPR_NEGATE ? -l : ~l);
         return true;
 
     case EXPR_UNARY_PLUS:
-        left = expr->value.child;
+        left = expr->unary.child;
         return ExprResolveIntegerLookup(ctx, left, val_rtrn, lookup,
                                         lookupPriv);
 
     default:
-        log_wsgo(ctx, "Unknown operator %d in ResolveInteger\n", expr->op);
+        log_wsgo(ctx, "Unknown operator %d in ResolveInteger\n",
+                 expr->expr.op);
         break;
     }
 
@@ -431,40 +429,34 @@ ExprResolveLevel(struct xkb_context *ctx, const ExprDef *expr,
 bool
 ExprResolveButton(struct xkb_context *ctx, const ExprDef *expr, int *btn_rtrn)
 {
-    int result;
-
-    if (!ExprResolveIntegerLookup(ctx, expr, &result, SimpleLookup,
-                                  buttonNames))
-        return false;
-
-    *btn_rtrn = result;
-    return true;
+    return ExprResolveIntegerLookup(ctx, expr, btn_rtrn, SimpleLookup,
+                                    buttonNames);
 }
 
 bool
 ExprResolveString(struct xkb_context *ctx, const ExprDef *expr,
                   xkb_atom_t *val_rtrn)
 {
-    switch (expr->op) {
+    switch (expr->expr.op) {
     case EXPR_VALUE:
-        if (expr->value_type != EXPR_TYPE_STRING) {
+        if (expr->expr.value_type != EXPR_TYPE_STRING) {
             log_err(ctx, "Found constant of type %s, expected a string\n",
-                    expr_value_type_to_string(expr->value_type));
+                    expr_value_type_to_string(expr->expr.value_type));
             return false;
         }
 
-        *val_rtrn = expr->value.str;
+        *val_rtrn = expr->string.str;
         return true;
 
     case EXPR_IDENT:
         log_err(ctx, "Identifier \"%s\" of type string not found\n",
-                xkb_atom_text(ctx, expr->value.str));
+                xkb_atom_text(ctx, expr->ident.ident));
         return false;
 
     case EXPR_FIELD_REF:
         log_err(ctx, "Default \"%s.%s\" of type string not found\n",
-                xkb_atom_text(ctx, expr->value.field.element),
-                xkb_atom_text(ctx, expr->value.field.field));
+                xkb_atom_text(ctx, expr->field_ref.element),
+                xkb_atom_text(ctx, expr->field_ref.field));
         return false;
 
     case EXPR_ADD:
@@ -477,11 +469,12 @@ ExprResolveString(struct xkb_context *ctx, const ExprDef *expr,
     case EXPR_NOT:
     case EXPR_UNARY_PLUS:
         log_err(ctx, "%s of strings not permitted\n",
-                expr_op_type_to_string(expr->op));
+                expr_op_type_to_string(expr->expr.op));
         return false;
 
     default:
-        log_wsgo(ctx, "Unknown operator %d in ResolveString\n", expr->op);
+        log_wsgo(ctx, "Unknown operator %d in ResolveString\n",
+                 expr->expr.op);
         break;
     }
     return false;
@@ -491,16 +484,16 @@ bool
 ExprResolveEnum(struct xkb_context *ctx, const ExprDef *expr,
                 unsigned int *val_rtrn, const LookupEntry *values)
 {
-    if (expr->op != EXPR_IDENT) {
+    if (expr->expr.op != EXPR_IDENT) {
         log_err(ctx, "Found a %s where an enumerated value was expected\n",
-                expr_op_type_to_string(expr->op));
+                expr_op_type_to_string(expr->expr.op));
         return false;
     }
 
-    if (!SimpleLookup(ctx, values, expr->value.str, EXPR_TYPE_INT,
+    if (!SimpleLookup(ctx, values, expr->ident.ident, EXPR_TYPE_INT,
                       val_rtrn)) {
         log_err(ctx, "Illegal identifier %s; expected one of:\n",
-                xkb_atom_text(ctx, expr->value.str));
+                xkb_atom_text(ctx, expr->ident.ident));
         while (values && values->name)
         {
             log_err(ctx, "\t%s\n", values->name);
@@ -523,29 +516,29 @@ ExprResolveMaskLookup(struct xkb_context *ctx, const ExprDef *expr,
     ExprDef *left, *right;
     const char *bogus = NULL;
 
-    switch (expr->op) {
+    switch (expr->expr.op) {
     case EXPR_VALUE:
-        if (expr->value_type != EXPR_TYPE_INT) {
+        if (expr->expr.value_type != EXPR_TYPE_INT) {
             log_err(ctx,
                     "Found constant of type %s where a mask was expected\n",
-                    expr_value_type_to_string(expr->value_type));
+                    expr_value_type_to_string(expr->expr.value_type));
             return false;
         }
-        *val_rtrn = (unsigned int) expr->value.ival;
+        *val_rtrn = (unsigned int) expr->integer.ival;
         return true;
 
     case EXPR_IDENT:
-        ok = lookup(ctx, lookupPriv, expr->value.str, EXPR_TYPE_INT,
+        ok = lookup(ctx, lookupPriv, expr->ident.ident, EXPR_TYPE_INT,
                     val_rtrn);
         if (!ok)
             log_err(ctx, "Identifier \"%s\" of type int is unknown\n",
-                    xkb_atom_text(ctx, expr->value.str));
+                    xkb_atom_text(ctx, expr->ident.ident));
         return ok;
 
     case EXPR_FIELD_REF:
         log_err(ctx, "Default \"%s.%s\" of type int is unknown\n",
-                xkb_atom_text(ctx, expr->value.field.element),
-                xkb_atom_text(ctx, expr->value.field.field));
+                xkb_atom_text(ctx, expr->field_ref.element),
+                xkb_atom_text(ctx, expr->field_ref.field));
         return false;
 
     case EXPR_ARRAY_REF:
@@ -563,13 +556,13 @@ ExprResolveMaskLookup(struct xkb_context *ctx, const ExprDef *expr,
     case EXPR_SUBTRACT:
     case EXPR_MULTIPLY:
     case EXPR_DIVIDE:
-        left = expr->value.binary.left;
-        right = expr->value.binary.right;
+        left = expr->binary.left;
+        right = expr->binary.right;
         if (!ExprResolveMaskLookup(ctx, left, &l, lookup, lookupPriv) ||
             !ExprResolveMaskLookup(ctx, right, &r, lookup, lookupPriv))
             return false;
 
-        switch (expr->op) {
+        switch (expr->expr.op) {
         case EXPR_ADD:
             *val_rtrn = l | r;
             break;
@@ -579,7 +572,7 @@ ExprResolveMaskLookup(struct xkb_context *ctx, const ExprDef *expr,
         case EXPR_MULTIPLY:
         case EXPR_DIVIDE:
             log_err(ctx, "Cannot %s masks; Illegal operation ignored\n",
-                    (expr->op == EXPR_DIVIDE ? "divide" : "multiply"));
+                    (expr->expr.op == EXPR_DIVIDE ? "divide" : "multiply"));
             return false;
         default:
             break;
@@ -592,7 +585,7 @@ ExprResolveMaskLookup(struct xkb_context *ctx, const ExprDef *expr,
         break;
 
     case EXPR_INVERT:
-        left = expr->value.child;
+        left = expr->unary.child;
         if (!ExprResolveIntegerLookup(ctx, left, &v, lookup, lookupPriv))
             return false;
 
@@ -602,14 +595,15 @@ ExprResolveMaskLookup(struct xkb_context *ctx, const ExprDef *expr,
     case EXPR_UNARY_PLUS:
     case EXPR_NEGATE:
     case EXPR_NOT:
-        left = expr->value.child;
+        left = expr->unary.child;
         if (!ExprResolveIntegerLookup(ctx, left, &v, lookup, lookupPriv))
             log_err(ctx, "The %s operator cannot be used with a mask\n",
-                    (expr->op == EXPR_NEGATE ? "-" : "!"));
+                    (expr->expr.op == EXPR_NEGATE ? "-" : "!"));
         return false;
 
     default:
-        log_wsgo(ctx, "Unknown operator %d in ResolveMask\n", expr->op);
+        log_wsgo(ctx, "Unknown operator %d in ResolveMask\n",
+                 expr->expr.op);
         break;
     }
 
@@ -624,12 +618,12 @@ ExprResolveMask(struct xkb_context *ctx, const ExprDef *expr,
 }
 
 bool
-ExprResolveModMask(struct xkb_keymap *keymap, const ExprDef *expr,
-                   enum mod_type mod_type, xkb_mod_mask_t *mask_rtrn)
+ExprResolveModMask(struct xkb_context *ctx, const ExprDef *expr,
+                   enum mod_type mod_type, const struct xkb_mod_set *mods,
+                   xkb_mod_mask_t *mask_rtrn)
 {
-    LookupModMaskPriv priv = { .keymap = keymap, .mod_type = mod_type };
-    return ExprResolveMaskLookup(keymap->ctx, expr, mask_rtrn, LookupModMask,
-                                 &priv);
+    LookupModMaskPriv priv = { .mods = mods, .mod_type = mod_type };
+    return ExprResolveMaskLookup(ctx, expr, mask_rtrn, LookupModMask, &priv);
 }
 
 bool
@@ -638,9 +632,8 @@ ExprResolveKeySym(struct xkb_context *ctx, const ExprDef *expr,
 {
     int val;
 
-    if (expr->op == EXPR_IDENT) {
-        const char *str;
-        str = xkb_atom_text(ctx, expr->value.str);
+    if (expr->expr.op == EXPR_IDENT) {
+        const char *str = xkb_atom_text(ctx, expr->ident.ident);
         *sym_rtrn = xkb_keysym_from_name(str, 0);
         if (*sym_rtrn != XKB_KEY_NoSymbol)
             return true;
@@ -652,31 +645,33 @@ ExprResolveKeySym(struct xkb_context *ctx, const ExprDef *expr,
     if (val < 0 || val >= 10)
         return false;
 
-    *sym_rtrn = ((xkb_keysym_t) val) + '0';
+    *sym_rtrn = XKB_KEY_0 + (xkb_keysym_t) val;
     return true;
 }
 
 bool
-ExprResolveMod(struct xkb_keymap *keymap, const ExprDef *def,
-               enum mod_type mod_type, xkb_mod_index_t *ndx_rtrn)
+ExprResolveMod(struct xkb_context *ctx, const ExprDef *def,
+               enum mod_type mod_type, const struct xkb_mod_set *mods,
+               xkb_mod_index_t *ndx_rtrn)
 {
     xkb_mod_index_t ndx;
-    xkb_atom_t name = def->value.str;
+    xkb_atom_t name;
 
-    if (def->op != EXPR_IDENT) {
-        log_err(keymap->ctx,
+    if (def->expr.op != EXPR_IDENT) {
+        log_err(ctx,
                 "Cannot resolve virtual modifier: "
                 "found %s where a virtual modifier name was expected\n",
-                expr_op_type_to_string(def->op));
+                expr_op_type_to_string(def->expr.op));
         return false;
     }
 
-    ndx = ModNameToIndex(keymap, name, mod_type);
+    name = def->ident.ident;
+    ndx = XkbModNameToIndex(mods, name, mod_type);
     if (ndx == XKB_MOD_INVALID) {
-        log_err(keymap->ctx,
+        log_err(ctx,
                 "Cannot resolve virtual modifier: "
                 "\"%s\" was not previously declared\n",
-                xkb_atom_text(keymap->ctx, name));
+                xkb_atom_text(ctx, name));
         return false;
     }
 
